/* Hashao Product.
   Version: 0.1.1
   Convert a web page between Simplified Chinese and Traditional Chinese.
 */

var FRAMEDEEP = 18; // Recursive frame/iframe deep.
var zh_chars = { // Possible Chinese charsets.
	"cn": {"GB2312":1, "x-gbk":1, "gb18030":1, 
	       "HZ-GB-2312":1, "ISO-2022-CN":1},
	"tw": {"Big5":1, "x-euc-tw":1},
	"hk": {"Big5-HKSCS":1}
    };
/* Possible charset that has Chinese content. */
var zh_all = {"GB2312":1, "x-gbk":1, "gb18030":1, 
              "HZ-GB-2312":1, "ISO-2022-CN":1,
              "Big5":1, "x-euc-tw":1, "Big5-HKSCS":1,
	      "UTF-7":1, "UTF-8":1, "UTF-16LE":1,
	      "x-user-defined":1};
var SIMFLAG = 's';
var TRAFLAG = 't';
var DEBUG = false;

const notWhiteSpace = /\S/; // Match for non-white-space text.

/* Get the tongwen flag of a root node.
 */
function getZhFlag(curDoc, root) {
    try {
	// Attach a zhongwen attribute to the root so we can switch zh styles.
	var zhflag = null;
	zhflag = root.getAttribute("zhongwen");
    } catch(e) {
	alert("getZhFlag " + e + ": " + "curDoc" );
    }
    return zhflag;
}

/* Guess the tongwen flag of a Doc. */
function guessZhFlag(curDoc, root) {
    var zhflag = getZhFlag(curDoc, root);
    if (! zhflag) {
	var charset = curDoc.characterSet;
	if (charset in zh_chars["tw"] || charset in zh_chars["hk"]) {
	    zhflag = TRAFLAG;
	}
	else
	    zhflag = SIMFLAG;
	setZhFlag(root, zhflag);
    }
    return zhflag;
}

/* Set the tongwen flag of a Doc. */
function setZhFlag(root, zhflag) {
    root.setAttribute("zhongwen", zhflag);
}

/* Make a doc to switch language or force it to one language. */
function tongWen(forcezh) {
    try {
	var curDoc = getFocusedDoc();
	var zhflag = null;
	// If no HTMLDocument, we just stop.
	if (!curDoc || curDoc instanceof XULDocument)
	    return;

	if (forcezh) {
	    zhflag = forcezh;
	} else {
	    var oldflag = guessZhFlag(curDoc, curDoc.body);
	    zhflag = (oldflag != SIMFLAG)? SIMFLAG:TRAFLAG;
	}

	doFrames(curDoc, zhflag, 0);

    } catch(e) { alert("tongWen " + e + ": " + curDoc); }
}

/* Do a list of frames, have to be recursive someohow, whatever.. */
function doFrames(curDoc, zhflag, deep)
{
    try {
	// At least do current doc once.
	doADoc(curDoc, zhflag);
	deep++;
	var my_iframes = curDoc.getElementsByTagName("IFRAME");
	if (my_iframes.length != 0 && deep < FRAMEDEEP) {
	    for (var i = 0; i < my_iframes.length; i++) {
		curDoc = my_iframes[i].contentDocument;
		doFrames(curDoc, zhflag, deep);
	    }
	}
	var my_frames = curDoc.getElementsByTagName("FRAME");
	if (my_frames.length != 0 && deep < FRAMEDEEP) {
	    for (i = 0; i < my_frames.length; i++) {
		curDoc = my_frames[i].contentDocument;
		doFrames(curDoc, zhflag, deep);
	    }
	}
    } catch(e) {
	alert("doFrameLs " + e);
    }
}

/* Do a single document, then set flags. */
function doADoc(curDoc, zhflag)
{
    try {
	var curChar = curDoc.characterSet;
	var curflag = getZhFlag(curDoc, curDoc.body);

	/* already converted. */
	if (DEBUG) {
	    var msg = "In doADoc(), current state: ";
	    msg += '\ncurDoc title: ' + curDoc.title;
	    msg += '\ncurflag: ' + curflag + ', zflag: ' + zhflag;
	    msg += '\nbody.lang: ' + curDoc.body.lang;
	    alert(msg);
	}
	if (curflag && curflag == zhflag)
	    return;

	/* No chinese. OK, don't think we need this either. */
	if (!(curflag ||(curChar in zh_all))) {
	    /* Should just return but some shitty pages don't set proper encoding
	     * and use &#nnnnn; to represent Chinese, duh!
	     */
	    if (!curDoc.body.lang) {
		curDoc.body.lang = "zh-cn";
	    }
	    if (curDoc.body.lang == "zh-cn")
		curflag = SIMFLAG;
	    else
		curflag = TRAFLAG;
	    // return;
	}
	/* Dead code. We don't need code here.
	else {
	    curflag = guessZhFlag(curDoc, curDoc.body);
	}
	*/

	// Set lang attr. This is needed to force proper fonts.
	// Will not handle subnodes lang, but hey... 
	if (zhflag == TRAFLAG) {
	    /* For tradition charsets, single out hk language in case
	       mozilla add hk font option in the future and I don't
	       want to update anything. Yes, hk has special glyphs. */
	    if (curChar in zh_chars["hk"])
		curDoc.body.lang="zh-hk";
	    else
		curDoc.body.lang="zh-tw";
	}
	else
	    curDoc.body.lang="zh-cn";

	convertTree(curDoc, curDoc.body, zhflag);
	curDoc.title = convZh(curDoc.title, zhflag);
	setZhFlag(curDoc.body, zhflag);
    } catch(e) {
	alert("doADoc " + e + curDoc);
    }
}

/* 
 * Walk over a Dom Node and do translations.
 */
function convertTree(curDoc, root, zhflag) {
    try {
	// if (root.nodeName == "BODY") {
	if (false) { /* Currently cannot be done with DOM. */
	    /* Try to optimize for 'body' node. */
	    convertDoc(curDoc, zhflag);
	}else {
	    var tw = curDoc.createTreeWalker(root, NodeFilter.SHOW_TEXT, 
		    null, true);
	    // Actual translation. want to do a recursive walk so in its own func.
	    proc_tree(tw, zhflag);
	}
    } catch(e) { alert("convertTree " + e); }
}

/* FIXME: Wrong code, dom cannot do this. */
/* Only work on a whole doc, not node. .getElementsByName is doc only. */
function convertDoc(curDoc, zhflag)
{
    try {
	var textnodes = curDoc.getElementsByTagName("#text");
	for (n in textnodes) {
	    var pname = n.parentNode.nodeName;
	    if (pname != "#SCRIPT" && pname != "#NOSCRIPT") {
		// txt += n.nodeValue; // Debug.
		if (notWhiteSpace.test(n.nodeValue))
		    n.nodeValue = convZh(n.nodeValue, zhflag);
	    }
	}
    } catch(e) { alert("convertDoc " + e); }
}

/* Not gonna do recursive cause too many levels.
 * Version before 0.1.1 walk full tree with SHOW_ALL.
 */
function proc_tree(tw, zhflag){
    var n = tw.currentNode; // OK, my node.
    var pname = null;
    // txt = "";

    // SHOW_TEXT only has 2 layers. parent->textnodes.
    if (n.nodeName != "#text")
	n = tw.firstChild(); 

    while (n) {
	pname = n.parentNode.nodeName;
	if (pname != "#SCRIPT" && pname != "#NOSCRIPT") {
	    // txt += n.nodeValue; // Debug.
	    // Skip empty nodes. Did not make much different.
	    //if (notWhiteSpace.test(n.nodeValue))
		n.nodeValue = convZh(n.nodeValue, zhflag);
	}
	n = tw.nextSibling();
    } // While(n)
    // alert(txt); // Debug.
}

/* Do simplified<-> traditional unicode conversion. 
   t_2_s/s_2_t are tables imported from t2s.js.
 */
function convZh(itxt, zhflag) {
    if (hasctrans) {
	/* If we have c xpcom module, take short cut. */
	var zhmap = null;
	if (zhflag == SIMFLAG)
	    zhmap = ct2s;
	else if (zhflag == TRAFLAG)
	    zhmap = cs2t;
	return zhmap.trans(itxt)
    }

    var zhmap = null; // chinese simplified<->traditional character map.
    if (zhflag == SIMFLAG)
	zhmap = t_2_s;
    else if (zhflag == TRAFLAG)
	zhmap = s_2_t;

    // Use an array to speed up immutable string operation.
    itxt = itxt.split("");

    for(var i = 0; i < itxt.length; i++) {
	/*
	if (itxt[i] in zhmap) 
	    txt += zhmap[itxt[i]];
	else
	    txt += itxt[i];
	*/
	if (itxt[i] in zhmap) {
	    itxt[i] = zhmap[itxt[i]];
	}
    }
    return itxt.join("");
}

/* Switch the debug flag. */
function debug_switch() {
    var item = document.getElementById("tongwen-debug-item");
    var c = item.getAttribute("checked");
    DEBUG = c;
}
